# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import scipy.sparse as sparse
import scipy.linalg as linalg
import numpy as np
import os
import matplotlib.pyplot as plt
import logging
import matplotlib.colors as colors
from allensdk.config.manifest import Manifest

def demix_time_dep_masks(raw_traces, stack, masks):
    '''

    :param raw_traces: extracted traces
    :param stack: movie (same length as traces)
    :param masks: binary roi masks
    :return: demixed traces
    '''
    N, T = raw_traces.shape
    _, x, y = masks.shape
    P = x * y

    if len(stack.shape) == 3:
        stack = stack.reshape(T, P)

    num_pixels_in_mask = np.sum(masks, axis=(1, 2))
    F = raw_traces.T * num_pixels_in_mask  # shape (T,N)
    F = F.T

    flat_masks = masks.reshape(N, P)
    flat_masks = sparse.csr_matrix(flat_masks)

    drop_frames = []
    demix_traces = np.zeros((N, T))

    for t in range(T):

        weighted_mask_sum = F[:, t]
        drop_test = (weighted_mask_sum == 0)

        if np.sum(drop_test == 0):
            norm_mat = sparse.diags(num_pixels_in_mask / weighted_mask_sum, offsets=0)
            stack_t = sparse.diags(stack[t], offsets=0)

            flat_weighted_masks = norm_mat.dot(flat_masks.dot(stack_t))

            overlap = flat_masks.dot(flat_weighted_masks.T).toarray()  # cast to dense numpy array for linear solver because solution is dense
            try:
                demix_traces[:, t] = linalg.solve(overlap, F[:, t])
            except linalg.LinAlgError as e:
                logging.warning("singular matrix, using least squares")
                x, _, _, _ = linalg.lstsq(overlap, F[:, t])
                demix_traces[:, t] = x

            drop_frames.append(False)

        else:
            drop_frames.append(True)

    return demix_traces, drop_frames

def plot_traces(raw_trace, demix_trace, roi_id, roi_ind, save_file):
    fig, ax = plt.subplots()

    ax.plot(raw_trace, label='Fluoresence')
    ax.plot(demix_trace, label='Demixed')
    ax.set_title("ROI ID(%d) index (%d)" % (roi_id, roi_ind))
    ax.legend()
    plt.savefig(save_file)
    plt.close(fig)

def find_zero_baselines(traces):
    means = traces.mean(axis=1)
    stds =  traces.std(axis=1)
    return np.where((means-stds) < 0)
    
def plot_negative_baselines(raw_traces, demix_traces, mask_array, roi_ids_mask, plot_dir, ext='png'):
    N, T = raw_traces.shape
    _, x, y = mask_array.shape

    logging.debug("finding negative baselines")
    neg_inds = find_negative_baselines(demix_traces)[0]
    
    overlap_inds = set()
    logging.debug("detected negative baselines: %s", str(neg_inds))
    for roi_ind in neg_inds:
        Manifest.safe_mkdir(plot_dir)

        save_file = os.path.join(plot_dir, str(roi_ids_mask[roi_ind]) + '_negative.' + ext)
        plot_traces(raw_traces[roi_ind], demix_traces[roi_ind], roi_ids_mask[roi_ind], roi_ind, save_file)

        ''' plot overlapping masks '''
        save_file = os.path.join(plot_dir, str(roi_ids_mask[roi_ind]) + '_negative_masks.' + ext)
        roi_overlap_inds = plot_overlap_masks_lengthOne(roi_ind, mask_array, save_file)

        overlap_inds.update(roi_overlap_inds)

    zero_inds = find_zero_baselines(demix_traces)[0]
    logging.debug("detected zero baselines: %s", str(zero_inds))
    overlap_inds.update(zero_inds)

    return list(overlap_inds)
        

    
        
def plot_negative_transients(raw_traces, demix_traces, valid_roi, mask_array, roi_ids_mask, plot_dir, ext='png'):

    N, T = raw_traces.shape
    _, x, y = mask_array.shape

    logging.debug("finding negative transients")
    trans_ind_list1 = [find_negative_transients_threshold(trace=demix_traces[n]) for n in range(N)]
    rois_with_trans1 = [i for i in range(N) if len(trans_ind_list1[i]) > 0]
    rois_with_trans = np.unique(rois_with_trans1)
    rois_with_trans = [r for r in rois_with_trans if len(trans_ind_list1[r][0]) > 0]

    logging.debug("plotting negative transients")

    flat_masks = mask_array.reshape(N, x*y)
    overlap = flat_masks.dot(flat_masks.T)
    overlap -= np.diag(np.diag(overlap))

    for roi_ind in rois_with_trans:

        ''' plot biggest negative transient of this roi '''
        trans_ind_list = trans_ind_list1[roi_ind]

        trans_ind_list = trans_ind_list[0]
        trans_list = []
        for i in trans_ind_list:
            if i > 100 and i < T - 100:
                trans_list.append(demix_traces[roi_ind, i - 100:i + 100])
            elif i > 100 and i >= T - 100:
                trans_list.append(demix_traces[roi_ind, i - 100:])
            else:
                trans_list.append(demix_traces[roi_ind, :i + 100])

        # trans_list = [demix_traces[roi_ind, i-100:i+100] for i in trans_ind_list if i > 100 and i < Nt]
        Ntrans = len(trans_list)
        biggest_trans = 0
        for i in range(1, Ntrans):
            if np.amin(trans_list[i]) < np.amin(trans_list[biggest_trans]):
                biggest_trans = i

        trans_ind = trans_ind_list[biggest_trans]

        # trans_ind_list = np.concatenate((trans_ind_list1[roi_ind][0], trans_ind_list2[roi_ind][0]))
        # trans_list_min = np.where(demix_traces[roi_ind, trans_ind_list] == min(demix_traces[roi_ind, trans_ind_list]))[0]

        if np.sum(overlap[roi_ind]) > 0:

            if valid_roi[roi_ind]:

                savefile = os.path.join(plot_dir, str(roi_ids_mask[roi_ind]) + '_transient_valid.' + ext)
                plot_transients(roi_ind, trans_ind, mask_array, raw_traces, demix_traces, savefile)

                ''' plot overlapping masks '''
                savefile = os.path.join(plot_dir, str(roi_ids_mask[roi_ind]) + '_masks_valid.' + ext)
                plot_overlap_masks_lengthOne(roi_ind, mask_array, savefile)
                # plot_overlap_masks(roi_ind, mask_test, savefile)
            else:
                savefile = os.path.join(plot_dir, str(roi_ids_mask[roi_ind]) + '_transient_invalid.' + ext)
                plot_transients(roi_ind, trans_ind, mask_array, raw_traces, demix_traces, savefile)

                ''' plot overlapping masks '''
                savefile = os.path.join(plot_dir, str(roi_ids_mask[roi_ind]) + '_masks_invalid.' + ext)
                plot_overlap_masks_lengthOne(roi_ind, mask_array, savefile)
                # plot_overlap_masks(roi_ind, mask_test, savefile)
                #
        else:
            continue

    return rois_with_trans


def rolling_window(trace, window=500):
    '''

    :param trace:
    :param window:
    :return:
    '''

    shape = trace.shape[:-1] + (trace.shape[-1] - window + 1, window)
    strides = trace.strides + (trace.strides[-1], )

    return np.lib.stride_tricks.as_strided(trace, shape=shape, strides=strides)


def find_negative_baselines(trace):
    means = trace.mean(axis=1)
    stds = trace.std(axis=1)
    return np.where((means+stds) < 0)

def find_negative_transients_threshold(trace, window=500, length=10, std_devs=3):
    trace = np.pad(trace, pad_width=(window-1, 0), mode='constant', constant_values=[np.mean(trace[:window])])
    rolling_mean = np.mean(rolling_window(trace, window), -1)
    rolling_std = np.std(rolling_window(trace, window), -1)

    below_thresh = (trace[window-1:] < rolling_mean - std_devs*rolling_std)
    below_thresh = np.pad(below_thresh, pad_width=(window-1, 0), mode='constant')
    trans_length = np.sum(rolling_window(below_thresh, length), -1)
    trans_length = trans_length[window-length:]

    trans_ind = np.where(trans_length == length)

    return trans_ind


def plot_overlap_masks_lengthOne(roi_ind, masks, savefile=None, weighted=False):

    masks = np.array(masks).astype(float)
    N, x, y = masks.shape
    if np.sum(masks[-1])  == x*y:
        masks = masks[:-1]
        N -= 1

    flat_masks = masks.reshape(N, x*y)
    masks_overlap = flat_masks.dot(flat_masks.T)

    ind_plot = np.where(masks_overlap[roi_ind, :] > 0)[0]  # rois (k) that roi_ind overlaps with
    for i in ind_plot:  # rois that overlap with each roi k
        ind_k = np.where(masks_overlap[i, :] > 0)[0]
        ind_plot = np.concatenate((ind_plot, ind_k))

    ind_plot = np.unique(ind_plot)
    ind_plot = np.concatenate(([roi_ind], ind_plot[ind_plot!=roi_ind]))

    plt.figure()
    color_list = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
    Ncol = len(color_list)
    for num, i in enumerate(ind_plot):
        mask_plot = masks[i]
        if not weighted:
            mask_plot = ((num % Ncol)+1)*np.ma.array(masks[i], mask=(masks[i] == 0))
            plt.imshow(mask_plot, clim=(1., Ncol+1), cmap=colors.ListedColormap(color_list), alpha=0.5, interpolation='nearest')
            # plt.imshow(mask_plot, clim=(1., len(ind_plot)), alpha=.5)

        elif weighted:
            mask_plot = np.ma.array(masks[i], mask=(masks[i] == 0))
            plt.imshow(mask_plot, cmap='gray_r', alpha=.5, interpolation='nearest')

        plt.text(np.mean(np.where(np.sum(mask_plot, axis=0))), np.mean(np.where(np.sum(mask_plot, axis=1))) ,str(i))

    mask_tot = np.sum(masks[ind_plot, :, :], axis=0)
    mask_x = np.sum(mask_tot, axis=0)
    mask_y = np.sum(mask_tot, axis=1)

    plt.xlim((np.amin(np.where(mask_x))-5, np.amax(np.where(mask_x))+5))
    plt.ylim((np.amin(np.where(mask_y))-5, np.amax(np.where(mask_y))+5))
    plt.title('Masks')

    if savefile is not None:
        plt.savefig(savefile)
    plt.close()

    return ind_plot


def plot_transients(roi_ind, t_trans, masks, traces, demix_traces, savefile):

    masks = np.array(masks).astype(float)
    N, x, y = masks.shape
    _, Nt = traces.shape

    flat_masks = masks.reshape(N, x*y)
    masks_overlap = flat_masks.dot(flat_masks.T)

    ind_plot = np.where(masks_overlap[roi_ind, :] > 0)[0]  # rois (k) that roi_ind overlaps with
    for i in ind_plot:  # rois that overlap with each roi k
        ind_k = np.where(masks_overlap[i, :] > 0)[0]
        ind_plot = np.concatenate((ind_plot, ind_k))

    ind_plot = np.unique(ind_plot)
    ind_plot = np.concatenate(([roi_ind], ind_plot[ind_plot!=roi_ind]))

    if t_trans > 150 and t_trans < Nt - 150:
        plot_t = range(t_trans - 150, t_trans + 150)
    elif t_trans > 150 and t_trans >= Nt - 150:
        plot_t = range(t_trans - 150, Nt)
    else:
        plot_t = range(0, t_trans + 150)

    fig, ax = plt.subplots(1, 2, figsize=(12, 6), sharex=True, sharey=True)
    color_list = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
    Ncol = len(color_list)

    for num, i in enumerate(ind_plot):
        ax[0].plot(plot_t, traces[i, plot_t], label=str(i), color=color_list[(num % Ncol)])
        ax[1].plot(plot_t, demix_traces[i, plot_t], label=str(i), color=color_list[(num % Ncol)])

    ax[0].set_title('Raw')
    ax[0].set_ylabel('Fluorescence')
    ax[1].set_title('Demixed')
    ax[1].set_xlabel('Time')
    ax[0].legend(loc=0)

    plt.savefig(savefile)
    plt.close(fig)




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
def list_of_dicts_to_dict_of_lists(list_of_dicts):
    return {key:[item[key] for item in list_of_dicts] for key in list_of_dicts[0].keys() }

def dict_generator(indict, pre=None):

    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in indict.items():
            if isinstance(value, dict):
                for d in dict_generator(value, pre + [key] ):
                    yield d
            elif isinstance(value, list):
                for v in value:
                    for d in dict_generator(v, pre + [key]):
                        yield d
            else:
                yield pre + [key, value]
    else:
        yield indict

def read_h5_group(g):
    return_dict = {}
    if len(g.attrs) > 0:
        return_dict['attrs'] = dict(g.attrs)
    for key in g:
        if key == 'data':
            return_dict[key] = g[key].value
        else:
            return_dict[key] = read_h5_group(g[key])

    return return_dict




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import functools


class RmaPager(object):
    def __init__(self):
        pass

    @staticmethod
    def pager(fn,
              *args,
              **kwargs):
        total_rows = kwargs.pop('total_rows', None)
        num_rows = kwargs.get('num_rows', None)

        if total_rows == 'all':
            start_row = 0
            result_count = num_rows
            kwargs = kwargs
            kwargs['count'] = False

            while result_count == num_rows:
                kwargs['start_row'] = start_row
                data = fn(*args, **kwargs)
                start_row = start_row + num_rows
                result_count = len(data)
                for r in data:
                    yield r
        else:
            start_row = 0
            kwargs = kwargs
            kwargs['count'] = False

            while start_row < total_rows:
                kwargs['start_row'] = start_row

                data = fn(*args, **kwargs)
                result_count = len(data)
                start_row = start_row + result_count
                for r in data:
                    yield r

def pageable(total_rows=None,
             num_rows=None):
    def decor(func):
        decor.total_rows=total_rows
        decor.num_rows=num_rows

        @functools.wraps(func)
        def w(*args,
              **kwargs):
            if decor.num_rows and not 'num_rows' in kwargs:
                kwargs['num_rows'] = decor.num_rows
            if decor.total_rows and not 'total_rows' in kwargs:
                kwargs['total_rows'] = decor.total_rows

            result = RmaPager.pager(func,
                                    *args,
                                    **kwargs)
            return result
        return w
    return decor



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import math
import scipy.ndimage.morphology as morphology
import logging
import h5py

# constants used for accessing border array
RIGHT_SHIFT = 0
LEFT_SHIFT = 1
DOWN_SHIFT = 2
UP_SHIFT = 3


class Mask(object):
    '''
    Abstract class to represent image segmentation mask. Its two
    main subclasses are RoiMask and NeuropilMask. The former represents
    the mask of a region of interest (ROI), such as a cell observed in
    2-photon imaging. The latter represents the neuropil around that cell,
    and is useful when subtracting the neuropil signal from the measured
    ROI signal.

    This class should not be instantiated directly.

    Parameters
    ----------
    image_w: integer
       Width of image that ROI resides in

    image_h: integer
       Height of image that ROI resides in

    label: text
       User-defined text label to identify mask

    mask_group: integer
       User-defined number to help put masks into different categories
    '''

    def __init__(self, image_w, image_h, label, mask_group):
        '''
        Mask class constructor. The Mask class is designed to be abstract
        and it should not be instantiated directly.
        '''

        self.img_rows = image_h
        self.img_cols = image_w
        # initialize to invalid state. Mask must be manually initialized
        #   by pixel list or mask array
        self.x = 0
        self.width = 0
        self.y = 0
        self.height = 0
        self.mask = None
        self.overlaps_motion_border = False
        # label is for distinguishing neuropil from ROI, in case
        #   these masks are mixed together
        self.label = label
        # auxiliary metadata. if a particula mask is part of an group,
        #   that data can be stored here
        self.mask_group = mask_group

    def __str__(self):
        return "%s: TL=%d,%d w,h=%d,%d\n%s" % (self.label, self.x, self.y, self.width, self.height, str(self.mask))

    def init_by_pixels(self, border, pix_list):
        '''
        Initialize mask using a list of mask pixels

        Parameters
        ----------
        border: float[4]
            Coordinates defining useable area of image. See create_roi_mask()

        pix_list: integer[][2]
            List of pixel coordinates (x,y) that define the mask
        '''
        assert pix_list.shape[1] == 2, "Pixel list not properly formed"
        array = np.zeros((self.img_rows, self.img_cols), dtype=bool)

        # pix_list stores array of [x,y] coordinates
        array[pix_list[:, 1], pix_list[:, 0]] = 1

        self.init_by_mask(border, array)

    def get_mask_plane(self):
        '''
        Returns mask content on full-size image plane

        Returns
        -------
        numpy 2D array [img_rows][img_cols]
        '''
        mask = np.zeros((self.img_rows, self.img_cols))
        mask[self.y:self.y + self.height, self.x:self.x + self.width] = self.mask
        return mask


def create_roi_mask(image_w, image_h, border, pix_list=None, roi_mask=None, label=None, mask_group=-1):
    '''
    Conveninece function to create and initializes an RoiMask

    Parameters
    ----------

    image_w: integer
        Width of image that ROI resides in

    image_h: integer
        Height of image that ROI resides in

    border: float[4]
        Coordinates defining useable area of image. If the entire image
        is usable, and masks are valid anywhere in the image, this should
        be [(image_w-1), 0, (image_h-1), 0]. The following constants
        help describe the array order:

            RIGHT_SHIFT = 0

            LEFT_SHIFT = 1

            DOWN_SHIFT = 2

            UP_SHIFT = 3

        When parts of the image are unusable, for example due motion
        correction shifting of different image frames, the border array
        should store the usable image area

    pix_list: integer[][2]
        List of pixel coordinates (x,y) that define the mask

    roi_mask: integer[image_h][image_w]
        Image-sized array that describes the mask. Active parts of the
        mask should have values >0. Background pixels must be zero

    label: text
        User-defined text label to identify mask

    mask_group: integer
        User-defined number to help put masks into different categories

    Returns
    -------
        RoiMask object
    '''
    m = RoiMask(image_w, image_h, label, mask_group)
    if pix_list is not None:
        m.init_by_pixels(border, pix_list)
    elif roi_mask is not None:
        m.init_by_mask(border, roi_mask)
    else:
        assert False, "Must specify either roi_mask or pix_list"
    return m


class RoiMask(Mask):

    def __init__(self, image_w, image_h, label, mask_group):
        '''
        RoiMask class constructor

        Parameters
        ----------
        image_w: integer
            Width of image that ROI resides in

        image_h: integer
            Height of image that ROI resides in

        label: text
            User-defined text label to identify mask

        mask_group: integer
            User-defined number to help put masks into different categories
        '''
        super(RoiMask, self).__init__(image_w, image_h, label, mask_group)

    def init_by_mask(self, border, array):
        '''
        Initialize mask using spatial mask

        Parameters
        ----------
        border: float[4]
            Coordinates defining useable area of image. See create_roi_mask().

        roi_mask: integer[image height][image width]
            Image-sized array that describes the mask. Active parts of the
            mask should have values >0. Background pixels must be zero
        '''
        # find lowest and highest non-zero indices on each axis
        px = np.argwhere(array)
        (top, left), (bottom, right) = px.min(0), px.max(0)

        # left and right border insets
        l_inset = math.ceil(border[RIGHT_SHIFT])
        r_inset = math.floor(self.img_cols - border[LEFT_SHIFT]) - 1
        # top and bottom border insets
        t_inset = math.ceil(border[DOWN_SHIFT])
        b_inset = math.floor(self.img_rows - border[UP_SHIFT]) - 1

        # if ROI crosses border, it's considered invalid
        if left < l_inset or right > r_inset:
            self.overlaps_motion_border = True
        if top < t_inset or bottom > b_inset:
            self.overlaps_motion_border = True
        #
        self.x = left
        self.width = right - left + 1
        self.y = top
        self.height = bottom - top + 1
        # make copy of mask
        self.mask = array[top:bottom + 1, left:right + 1]


def create_neuropil_mask(roi, border, combined_binary_mask, label=None):
    '''
    Conveninece function to create and initializes a Neuropil mask.
    Neuropil masks are defined as the region around an ROI, up to 13
    pixels out, that does not include other ROIs

    Parameters
    ----------

    roi: RoiMask object
        The ROI that the neuropil masks will be based on

    border: float[4]
        Coordinates defining useable area of image. See create_roi_mask().

    combined_binary_mask
        List of pixel coordinates (x,y) that define the mask

    combined_binary_mask: integer[image_h][image_w]
        Image-sized array that shows the position of all ROIs in the
        image. ROI masks should have a value of one. Background pixels
        must be zero. In other words, ithe combined_binary_mask is a
        bitmap union of all ROI masks

    label: text
        User-defined text label to identify the mask

    Returns
    -------
        NeuropilMask object
    '''
    # combined_binary_mask is a bitmap union of ALL ROI masks
    # create a binary mask of the ROI
    binary_mask = np.zeros((roi.img_rows, roi.img_cols))
    binary_mask[roi.y:roi.y + roi.height, roi.x:roi.x + roi.width] = roi.mask
    binary_mask = binary_mask > 0
    # dilate the mask
    binary_mask_dilated = morphology.binary_dilation(
        binary_mask, structure=np.ones((3, 3)), iterations=13)  # T/F
    # eliminate ROIs from the dilation
    binary_mask_dilated = binary_mask_dilated > combined_binary_mask
    # create mask from binary dilation
    m = NeuropilMask(w=roi.img_cols, h=roi.img_rows,
                     label=label, mask_group=roi.mask_group)
    m.init_by_mask(border, binary_mask_dilated)
    return m


class NeuropilMask(Mask):

    def __init__(self, w, h, label, mask_group):
        '''
        NeuropilMask class constructor. This class should be created by
        calling create_neuropil_mask()

        Parameters
        ----------
        label: text
            User-defined text label to identify mask

        mask_group: integer
            User-defined number to help put masks into different categories
        '''
        super(NeuropilMask, self).__init__(w, h, label, mask_group)

    def init_by_mask(self, border, array):
        '''
        Initialize mask using spatial mask

        Parameters
        ----------
        border: float[4]
            Coordinates defining useable area of image. See create_roi_mask().

        array: integer[image height][image width]
            Image-sized array that describes the mask. Active parts of the
            mask should have values >0. Background pixels must be zero
        '''
        # find lowest and highest non-zero indices on each axis
        px = np.argwhere(array)
        (top, left), (bottom, right) = px.min(0), px.max(0)

        # left and right border insets
        l_inset = math.ceil(border[RIGHT_SHIFT])
        r_inset = math.floor(self.img_cols - border[LEFT_SHIFT]) - 1
        # top and bottom border insets
        t_inset = math.ceil(border[DOWN_SHIFT])
        b_inset = math.floor(self.img_rows - border[UP_SHIFT]) - 1
        # restrict neuropil masks to center area of frame (ie, exclude
        #   areas that overlap with movement correction buffer)
        if left < l_inset:
            left = l_inset
            if right < l_inset:
                right = l_inset
        if right > r_inset:
            right = r_inset
            if left > r_inset:
                left = r_inset
        if top < t_inset:
            top = t_inset
            if bottom < t_inset:
                bottom = t_inset
        if bottom > b_inset:
            bottom = b_inset
            if top > b_inset:
                top = b_inset
        #
        self.x = left
        self.width = right - left + 1
        self.y = top
        self.height = bottom - top + 1
        # make copy of mask
        self.mask = array[top:bottom + 1, left:right + 1]


def calculate_traces(stack, mask_list, block_size=100):
    '''
    Calculates the average response of the specified masks in the
    image stack

    Parameters
    ----------
    stack: float[image height][image width]
        Image stack that masks are applied to

    mask_list: list<Mask>
        List of masks

    Returns
    -------
    float[number masks][number frames]
        This is the average response for each Mask in each image frame
    '''
    traces = np.zeros((len(mask_list), stack.shape[0]), dtype=float)
    num_frames = stack.shape[0]

    # make sure masks are numpy objects
    mask_areas = np.zeros(len(mask_list), dtype=float)
    valid_masks = np.ones(len(mask_list), dtype=bool)

    for i,mask in enumerate(mask_list):
        if not isinstance(mask.mask, np.ndarray):
            mask.mask = np.array(mask.mask)

        # compute mask areas
        mask_areas[i] = mask.mask.sum()

        # if the mask is empty, the trace is nan
        if mask_areas[i] == 0:
            logging.warning("mask '%d/%s' is empty", i, mask.label)
            traces[i,:] = np.nan
            valid_masks[i] = False

        # if the mask overlaps the motion border, the trace is nan
        if mask.overlaps_motion_border:
            logging.warning("mask '%d/%s' overlaps with motion border", i, mask.label)
            traces[i,:] = np.nan
            valid_masks[i] = False

    # calculate traces
    for frame_num in range(0, num_frames, block_size):
        if frame_num % 1000 == 0:
            logging.debug("frame " + str(frame_num) + " of " + str(num_frames))
        frames = stack[frame_num:frame_num+block_size]

        for i in range(len(mask_list)):
            if valid_masks[i] is False:
                continue

            mask = mask_list[i]
            subframe = frames[:,mask.y:mask.y + mask.height, 
                                mask.x:mask.x + mask.width]

            total = subframe[:, mask.mask].sum(axis=1)
            traces[i, frame_num:frame_num+block_size] = total / mask_areas[i]
    return traces

def calculate_roi_and_neuropil_traces(movie_h5, roi_mask_list, motion_border):
    """ get roi and neuropil masks """

    # a combined binary mask for all ROIs (this is used to 
    #   subtracted ROIs from annuli
    mask_array = create_roi_mask_array(roi_mask_list)
    combined_mask = mask_array.max(axis=0)

    logging.info("%d total ROIs" % len(roi_mask_list))

    # create neuropil masks for the central ROIs
    neuropil_masks = []
    for m in roi_mask_list:
        nmask = create_neuropil_mask(m, motion_border, combined_mask, "neuropil for " + m.label)
        neuropil_masks.append(nmask)

    # calculate fluorescence traces for valid ROI and neuropil masks
    # create a combined list and calculate these together (this lets us
    #   read the large image stack only once)
    combined_list = []
    for m in roi_mask_list:
        combined_list.append(m)
    for n in neuropil_masks:
        combined_list.append(n)

    with h5py.File(movie_h5, "r") as movie_f:
        stack_frames = movie_f["data"]

        logging.info("Calculating %d traces (neuropil + ROI) over %d frames" % (len(combined_list), len(stack_frames)))
        traces = calculate_traces(stack_frames, combined_list)

        roi_traces = traces[:len(roi_mask_list)]
        neuropil_traces = traces[len(roi_mask_list):]

    return roi_traces, neuropil_traces



def create_roi_mask_array(rois):
    '''Create full image mask array from list of RoiMasks.

    Parameters
    ----------
    rois: list<RoiMask>
        List of roi masks.

    Returns
    -------
    np.ndarray: NxWxH array
        Boolean array of of len(rois) image masks.
    '''
    if rois:
        height = rois[0].img_rows
        width = rois[0].img_cols
        masks = np.zeros((len(rois), height, width), dtype=np.uint8)
        for i, roi in enumerate(rois):
            masks[i, :, :] = roi.get_mask_plane()
    else:
        masks = None
    return masks




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import mock
from numpy import allclose

from allensdk.core.simple_tree import SimpleTree

@pytest.fixture
def tree():

    nodes = [{'id': 0, 'parent': None}, {'id': 1, 'parent': 0}, 
             {'id': 2, 'parent': 0}, {'id': 3, 'parent': 1}, 
             {'id': 4, 'parent': 1}, {'id': 5, 'parent': 2}]
            
    parent_fn = lambda node: node['parent']
    id_fn = lambda node: node['id']
    
    return SimpleTree(nodes, id_fn, parent_fn)
    
    
def test_initialization(tree):

    assert( None in tree._parent_ids.values() )
    assert( len(tree._child_ids) == 6 )
    
    
def test_filter_nodes(tree):
    
    two_par = tree.filter_nodes(lambda node: node['parent'] == 2)
    assert( two_par[0]['id'] == 5 )
    assert( len(two_par) == 1 )
    
    
def test_value_map(tree):
    
    parent_map = tree.value_map(lambda node: node['id'], 
                                lambda node: node['parent'])
                                
    assert( len(parent_map) == 6 )
    assert( parent_map[2] == 0 )
    assert( parent_map[3] == 1 ) 
    
    
def test_node_ids(tree):

    obtained = tree.node_ids()
    expected = range(6)
     
    assert( set(obtained) == set(expected) )  
    
    
def test_parent_id(tree):

    nodes = [5, 4, 2]
    obtained = tree.parent_id(nodes)
    
    assert( allclose([2, 1, 0], obtained) )
    
    
def test_child_ids(tree):

    obtained = tree.child_ids([1])
    assert( set(obtained[0]) == set([4, 3]) )
    assert( len(obtained) == 1 )
    
    
def test_ancestor_ids(tree):

    obtained = tree.ancestor_ids([5, 1])
    
    assert( len(obtained) == 2 )
    assert( set(obtained[0]) == set([5, 2, 0]) )
    assert( set(obtained[1]) == set([1, 0]) )
    
    
def test_descendant_ids(tree):

    obtained = tree.descendant_ids([0, 3])

    assert( len(obtained) == 2 )
    assert( set(obtained[0]) == set(range(6)) )
    assert( set(obtained[1]) == set([3]) )
    
    
def test_node(tree):
    
    obtained = tree.node([0, 1])
    
    assert( len(obtained) == 2 )
    assert( obtained[0]['parent'] is None )
    assert( obtained[1]['id'] == 1 )
    
    
def test_node_default(tree):

    obtained = tree.node()
    assert( len(obtained) == 6 )
    

def test_parent(tree):

    obtained = tree.parent([0, 1])
    assert( len(obtained) == 2 )
    assert( obtained[0] is None )

def test_children(tree):
    
    obtained = tree.children([0, 5])

    assert( len(obtained) == 2 )
    assert( set(obtained[1]) == set([]) )
    assert( len(obtained[0]) == 2 ) 
    assert( isinstance(obtained[0][0], dict) )
    
def test_descendants(tree):

    obtained = tree.descendants([0, 3])

    assert( len(obtained) == 2 )
    assert( len(obtained[0]) == 6 )
    assert( obtained[1][0]['id'] == 3 )
    assert( isinstance(obtained[0][0], dict) )


def test_ancestors(tree):

    obtained = tree.ancestors([5, 1])
    
    assert( len(obtained) == 2 )
    assert( len(obtained[0]) == 3 )
    assert( isinstance(obtained[0][0], dict) )
    assert( len(obtained[1]) == 2 )



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from .rma_api import RmaApi
from six import string_types


class ImageDownloadApi(RmaApi):
    '''HTTP Client to download whole or partial two-dimensional images from the Allen Institute
    with the SectionImage, AtlasImage and ProjectionImage Download Services.

    See `Downloading an Image <http://help.brain-map.org/display/api/Downloading+an+Image>`_
    for more documentation.
    '''

    _FILTER_TYPES = [ 'range', 'rgb', 'contrast' ]
    COLORMAPS = { "gray": 0,
                  "hotmetal": 1,
                  "jet": 2,
                  "redtemp": 3,
                  "expression": 4,
                  "red": 5,
                  "blue": 6,
                  "green": 7,
                  "aba": 8,
                  "aibsmap_alt": 9,
                  "colormap": 10,
                  "projection": 11
    }

    def __init__(self, base_uri=None):
        super(ImageDownloadApi, self).__init__(base_uri)

    def download_section_image(self,
                               section_image_id,
                               file_path=None,
                               **kwargs):
        self.download_image(section_image_id,
                            file_path,
                            endpoint=self.section_image_download_endpoint,
                            **kwargs)

    def download_atlas_image(self,
                             atlas_image_id,
                             file_path=None,
                             **kwargs):
        self.download_image(atlas_image_id,
                            file_path,
                            endpoint=self.atlas_image_download_endpoint,
                            **kwargs)

    def download_projection_image(self,
                                  projection_image_id,
                                  file_path=None,
                                  **kwargs):
        self.download_image(projection_image_id,
                            file_path,
                            endpoint=self.projection_image_download_endpoint,
                            **kwargs)

    def download_image(self,
                       image_id,
                       file_path=None,
                       endpoint=None,
                       **kwargs):
        ''' Download whole or partial two-dimensional images
        from the Allen Institute with the SectionImage or AtlasImage service.

        Parameters
        ----------
        image_id : integer
            SubImage to download.
        file_path : string, optional
            where to put it, defaults to image_id.jpg
        downsample : int, optional
            Number of times to downsample the original image.
        quality : int, optional
            jpeg quality of the returned image, 0 to 100 (default)
        expression : boolean, optional
            Request the expression mask for the SectionImage.
        view : string, optional
            'expression', 'projection', 'tumor_feature_annotation'
            or 'tumor_feature_boundary'
        top : int, optional
            Index of the topmost row of the region of interest.
        left :int, optional
            Index of the leftmost column of the region of interest.
        width : int, optional
            Number of columns in the output image.
        height : int, optional
            Number of rows in the output image.
        range : list of ints, optional
            Filter to specify the RGB channels. low,high,low,high,low,high
        colormap : list of floats, optional
            Filter to specify the RGB channels. [lower_threshold,colormap]
            gain 0-1, colormap id is a string from ImageDownloadApi.COLORMAPS
        rgb : list of floats, optional
            Filter to specify the RGB channels. [red,green,blue] 0-1
        contrast : list of floats, optional
            Filter to specify contrast parameters. [gain,bias] 0-1
        annotation : boolean, optional
            Request the annotated AtlasImage
        atlas : int, optional
            Specify the desired Atlas' annotations.
        projection : boolean, optional
            Request projection for the specified image.
        downsample_dimensions : boolean, optional
            Indicates if the width and height should be adjusted
            to account for downsampling.

        Returns
        -------
        None
            the file is downloaded and saved to the path.

        Notes
        -----
        By default, an unfiltered full-sized image with the highest quality
        is returned as a download if no parameters are provided.

        'downsample=1' halves the number of pixels of the original image
        both horizontally and vertically.        range_list = kwargs.get('range', None)

        
        Specifying 'downsample=2' quarters the height and width values.

        Quality must be an integer from 0, for the lowest quality,
        up to as high as 100. If it is not specified,
        it defaults to the highest quality.

        Top is specified in full-resolution (largest tier) pixel coordinates.
        SectionImage.y is the default value.

        Left is specified in full-resolution (largest tier) pixel coordinates.
        SectionImage.x is the default value.

        Width is specified in tier-resolution (desired tier) pixel coordinates.
        SectionImage.width is the default value. It is automatically adjusted when downsampled.

        Height is specified in tier-resolution (desired tier) pixel coordinates.
        SectionImage.height is the default value. It is automatically adjusted when downsampled.

        The range parameter consists of 6 comma delimited integers
        that define the lower (0) and upper (4095) bound for each channel in red-green-blue order
        (i.e. "range=0,1500,0,1000,0,4095").
        The default range values can be determined by referring to the following fields
        on the Equalization model associated with the SectionDataSet:
        red_lower, red_uppper, green_lower, green_upper, blue_lower, blue_upper.
        For more information, see the
        `Image Controls <http://help.brain-map.org/display/mouseconnectivity/Projection#Projection-ImageControls>`_
        section of the Allen Mouse Brain Connectivity Atlas:
        `Projection Dataset <http://help.brain-map.org/display/mouseconnectivity/Projection>`_
        help topic.
        See: `Image Download Service `<http://help.brain-map.org/display/api/Downloading+an+Image>_
        '''
        params = []

        if endpoint is None:
            endpoint = self.image_download_endpoint

        downsample = kwargs.get('downsample', None)

        if downsample is not None:
            params.append('downsample=%d' % (downsample))

        quality = kwargs.get('quality', None)

        if quality is not None:
            params.append('quality=%d' % (quality))

        tumor_feature_annotation = kwargs.get('tumor_feature_annotation', None)

        if tumor_feature_annotation is not None:
            if tumor_feature_annotation:
                params.append('tumor_feature_annotation=true')
            else:
                params.append('tumor_feature_annotation=false')

        tumor_feature_boundary = kwargs.get('tumor_feature_boundary', None)

        if tumor_feature_boundary is not None:
            if tumor_feature_boundary:
                params.append('tumor_feature_boundary=true')
            else:
                params.append('tumor_feature_boundary=false')

        annotation = kwargs.get('annotation', None)

        if annotation is not None:
            if annotation is True:
                params.append('annotation=true')
            else:
                params.append('annotation=false')

        atlas = kwargs.get('atlas', None)

        if atlas is not None:
            params.append('atlas=%d' % (atlas))

        projection = kwargs.get('projection', None)

        if projection is not None:
            if projection is True:
                params.append('projection=true')
            else:
                params.append('projection=false')

        expression = kwargs.get('expression', None)

        if expression is not None:
            if expression:
                params.append('expression=true')
            else:
                params.append('expression=false')

        colormap_filter = kwargs.get('colormap', None)
        
        if colormap_filter is not None:
            if isinstance(colormap_filter, string_types):
                params.append('colormap=%s' % (colormap_filter))
            else:
                lower_threshold = colormap_filter[0]
                colormap_id = ImageDownloadApi.COLORMAPS[colormap_filter[1]]
                filter_values_list = '0.5,%s,0,256,%d' % (str(lower_threshold),
                                                          colormap_id)
                params.append('colormap=%s' % (filter_values_list))

        # see
        # http://api.brain-map.org/api/v2/data/SectionDataSet/100141599.xml?include=equalization,section_images
        for filter_type in ImageDownloadApi._FILTER_TYPES:
            filter_values = kwargs.get(filter_type, None)
    
            if filter_values is not None:
                filter_values_list = ','.join(str(r) for r in filter_values)
                params.append('%s=%s' % (filter_type, filter_values_list))

        view = kwargs.get('view', None)

        if view is not None:
            if view in ['expression',
                        'projection',
                        'tumor_feature_annotation',
                        'tumor_feature_boundary']:
                params.append('view=%s' % (view))
            else:
                raise ValueError("view argument should be 'expression', 'projection', 'tumor_feature_annotation' or 'tumor_feature_boundary'")

        # region of interest
        for roi_key in ['left', 'top', 'width', 'height']:
            roi_value = kwargs.get(roi_key, None)
            if roi_value is not None:
                params.append('%s=%d' % (roi_key, roi_value))

        downsample_dimensions = kwargs.get('downsample_dimensions', None)

        if downsample_dimensions is not None:
            if downsample_dimensions:
                params.append('downsample_dimensions=true')
            else:
                params.append('downsample_dimensions=false')

        if len(params) > 0:
            url_params = "?" + "&".join(params)
        else:
            url_params = ''

        image_url = ''.join([endpoint,
                             '/',
                             str(image_id),
                             url_params])

        if file_path is None:
            file_path = '%d.jpg' % (image_id)

        self.retrieve_file_over_http(image_url, file_path)


    def atlas_image_query(self, atlas_id, image_type_name=None):
        '''Build the URL.

        Parameters
        ----------
        atlas_id : integer, optional
            request a certain record.
        image_type_name : string, optional
            if not present, the query will get it from the atlas id.

        Returns
        -------
        url : string
            The constructed URL

        Notes
        -----
        See `Downloading Atlas Images and Graphics <http://help.brain-map.org/display/api/Atlas+Drawings+and+Ontologies#AtlasDrawingsandOntologies-DownloadingAtlasImagesAndGraphics>`_
        for additional documentation.
        The atlas id can be found with :py:meth:`allensdk.api.queries.ontologies_api.OntologiesApi.build_atlases_query`
        '''
        rma = self

        stages = []

        if image_type_name is None:
            atlas_stage = rma.model_stage('Atlas',
                                          criteria='[id$eq%d]' % (atlas_id),
                                          only=['image_type'])
            stages.append(atlas_stage)

            atlas_name_pipe_stage = rma.pipe_stage('list',
                                                   parameters=[('type_name',
                                                                rma.IS,
                                                                rma.quote_string('image_type'))])
            stages.append(atlas_name_pipe_stage)

            image_type_name = '$type_name'
        else:
            image_type_name = rma.quote_string(image_type_name)

        criteria_list = ['[annotated$eqtrue],',
                         'atlas_data_set(atlases[id$eq%d]),' % (atlas_id),
                         "alternate_images[image_type$eq%s]" % (image_type_name)]

        atlas_image_model_stage = rma.model_stage('AtlasImage',
                                                  criteria=criteria_list,
                                                  order=[
                                                      'sub_images.section_number'],
                                                  num_rows='all')

        stages.append(atlas_image_model_stage)

        return self.json_msg_query(
            rma.build_query_url(stages))



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from allensdk.brain_observatory.natural_scenes import NaturalScenes
from allensdk.brain_observatory.stimulus_analysis import StimulusAnalysis
import pytest
from mock import patch, MagicMock


@pytest.fixture
def dataset():
    dataset = MagicMock(name='dataset')
    
    timestamps = MagicMock(name='timestamps')
    celltraces = MagicMock(name='celltraces')
    dataset.get_corrected_fluorescence_traces = \
        MagicMock(name='get_corrected_fluorescence_traces',
                  return_value=(timestamps, celltraces))
    dataset.get_roi_ids = MagicMock(name='get_roi_ids')
    dataset.get_cell_specimen_ids = MagicMock(name='get_cell_specimen_ids')
    dff_traces = MagicMock(name="dfftraces")
    dataset.get_dff_traces = MagicMock(name='get_dff_traces',
                                       return_value=(None, dff_traces))
    dxcm = MagicMock(name='dxcm')
    dxtime = MagicMock(name='dxtime')
    dataset.get_running_speed=MagicMock(name='get_running_speed',
                                        return_value=(dxcm, dxtime))
    dataset.get_stimulus_table=MagicMock(name='get_stimulus_table',
                                         return_value=MagicMock())
    
    return dataset

def mock_speed_tuning():
    binned_dx_sp = MagicMock(name='binned_dx_sp')
    binned_cells_sp = MagicMock(name='binned_cells_sp')
    binned_dx_vis = MagicMock(name='binned_dx_vis')
    binned_cells_vis = MagicMock(name='binned_cells_vis')
    peak_run = MagicMock(name='peak_run')
    
    return MagicMock(name='get_speed_tuning',
                     return_value=(binned_dx_sp,
                                   binned_cells_sp,
                                   binned_dx_vis,
                                   binned_cells_vis,
                                   peak_run))

def mock_sweep_response():
    sweep_response = MagicMock(name='sweep_response')
    mean_sweep_response = MagicMock(name='mean_sweep_response')
    pval = MagicMock(name='pval')
    
    return MagicMock(name='get_sweep_response',
                     return_value=(sweep_response,
                                   mean_sweep_response,
                                   pval))

@patch.object(StimulusAnalysis,
              'get_speed_tuning',
              mock_speed_tuning())
@patch.object(StimulusAnalysis,
              'get_sweep_response',
              mock_sweep_response())
@pytest.mark.parametrize('trigger', (1, 2, 3, 4, 5))
def test_harness(dataset, trigger):
    ns = NaturalScenes(dataset)

    assert ns._stim_table is StimulusAnalysis._PRELOAD
    assert ns._number_scenes is StimulusAnalysis._PRELOAD
    assert ns._sweeplength is StimulusAnalysis._PRELOAD
    assert ns._interlength is StimulusAnalysis._PRELOAD
    assert ns._extralength is StimulusAnalysis._PRELOAD
    assert ns._sweep_response is StimulusAnalysis._PRELOAD
    assert ns._mean_sweep_response is StimulusAnalysis._PRELOAD
    assert ns._pval is StimulusAnalysis._PRELOAD
    assert ns._response is StimulusAnalysis._PRELOAD
    assert ns._peak is StimulusAnalysis._PRELOAD

    if trigger == 1:
        print(ns._stim_table)
        print(ns.sweep_response)
        print(ns.response)
        print(ns.peak)
    if trigger == 2:
        print(ns.number_scenes)
        print(ns.sweep_response)
        print(ns.response)
        print(ns.peak)
    elif trigger == 3:
        print(ns.sweeplength)
        print(ns.mean_sweep_response)
        print(ns.response)
        print(ns.peak)
    elif trigger == 4:
        print(ns.interlength)
        print(ns.sweep_response)
        print(ns.response)
        print(ns.peak)
    elif trigger == 5:
        print(ns.extralength)
        print(ns.mean_sweep_response)
        print(ns.response)
        print(ns.peak)

    assert ns._stim_table is not StimulusAnalysis._PRELOAD
    assert ns._number_scenes is not StimulusAnalysis._PRELOAD
    assert ns._sweeplength is not StimulusAnalysis._PRELOAD
    assert ns._interlength is not StimulusAnalysis._PRELOAD
    assert ns._extralength is not StimulusAnalysis._PRELOAD
    assert ns._sweep_response is not StimulusAnalysis._PRELOAD
    assert ns._mean_sweep_response is not StimulusAnalysis._PRELOAD
    assert ns._pval is not StimulusAnalysis._PRELOAD
    assert ns._response is not StimulusAnalysis._PRELOAD
    assert ns._peak is not StimulusAnalysis._PRELOAD

    # check super properties
    dataset.get_corrected_fluorescence_traces.assert_called_once_with()
    assert ns._timestamps != NaturalScenes._PRELOAD
    assert ns._celltraces != NaturalScenes._PRELOAD
    assert ns._numbercells != NaturalScenes._PRELOAD

    assert not dataset.get_roi_ids.called
    assert ns._roi_id is NaturalScenes._PRELOAD

    assert dataset.get_cell_specimen_ids.called
    assert ns._cell_id is NaturalScenes._PRELOAD

    assert not dataset.get_dff_traces.called
    assert ns._dfftraces is NaturalScenes._PRELOAD

    assert ns._dxcm is NaturalScenes._PRELOAD
    assert ns._dxtime is NaturalScenes._PRELOAD



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import scipy.interpolate as si
import scipy.ndimage.filters as filt
import scipy.stats as stats

def chi_square_binary(events, LSN_template):
    # note: can only be applied to binary events for trial responses
    #
    # *****INPUT*****
    # events: 2D numpy bool with shape (num_trials,num_cells) for presence
    #   or absence of response on a given trial
    # LSN_template: 3D numpy int8 with shape (num_trials,num_y_pixels,num_x_pixels)
    #   for luminance at each pixel location
    #
    # *****OUTPUT*****
    # chi_square_grid_NLL: 3D numpy float with shape (num_cells,num_y_pixels,num_x_pixels)
    #   that gives the negative log-likelihood that a receptive field is contained
    #   within a 7x7 pixel mask centered on a given pixel location as measured
    #   by a chi-square test for the responses among the pixels (both on and off)
    #   that fall within the mask.

    num_trials = np.shape(events)[0]
    num_cells = np.shape(events)[1]
    num_y = np.shape(LSN_template)[1]
    num_x = np.shape(LSN_template)[2]

    # for each pixel location, get a mask that is centered on that location
    #    disc_masks has shape (num_y,num_x,num_y,num_x)
    disc_masks = get_disc_masks(LSN_template)

    # determine which trials each pixel is active (i.e not gray),
    #   broken up by ON and OFF pixels.
    #   trial_matrix has shape (num_y,num_x,2,num_trials)
    trial_matrix = build_trial_matrix(LSN_template, num_trials)

    # get the total number of trials each pixel is active (i.e. not gray)
    #   trials_per_pixel has shape (num_y,num_x,2)
    trials_per_pixel = np.sum(trial_matrix, axis=3)

    # get the sum of the number of events across all trials that each pixel is active
    #   events_per_pixel has shape (num_cells,num_y,num_x,2)
    events_per_pixel = get_events_per_pixel(events, trial_matrix)

    # smooth stimulus-triggered average spatially with a gaussian
    for n in range(num_cells):
        for on_off in range(2):
            events_per_pixel[n, :, :, on_off] = smooth_STA(events_per_pixel[n, :, :, on_off])

    # calculate the p_value for each exclusion region
    chi_square_grid = np.zeros((num_cells, num_y, num_x))
    for y in range(num_y):
        for x in range(num_x):
            exclusion_mask = np.ones((num_y, num_x, 2)) * disc_masks[y, x, :, :].reshape(num_y, num_x, 1)
            p_vals, __ = chi_square_within_mask(exclusion_mask, events_per_pixel, trials_per_pixel)
            chi_square_grid[:, y, x] = p_vals

    # convert p-values to negative log-likelihood
    # chi_square_grid_NLL = pvalue_to_NLL(chi_square_grid)

    return chi_square_grid



def get_peak_significance(chi_square_grid_NLL,
                          LSN_template,
                          alpha=0.05):
    # ****INPUT*****
    # chi_square_grid_NLL: result of chi_square_binary(events,LSN_template)
    # LSN_template: 3D numpy int8 with shape (num_trials,num_y_pixels,num_x_pixels)
    #   for luminance at each pixel location
    #
    # *****OUTPUT*****
    # significant_cells: 1D numpy bool with shape (num_cells,) that indicates
    #   whether or not each cell has a location with a significant chance of a
    #   true receptive field
    # best_exclusion_region_list: list of 2D numpy bool with len (num_cells). For each cell,
    #   the array is a mask for the best pixel, or all false

    num_cells = np.shape(chi_square_grid_NLL)[0]
    num_y = np.shape(chi_square_grid_NLL)[1]
    num_x = np.shape(chi_square_grid_NLL)[2]

    chi_square_grid = NLL_to_pvalue(chi_square_grid_NLL)

    # get the average size of all masks in units of number of pixels
    disc_masks = get_disc_masks(LSN_template)
    pixels_per_mask_per_pixel = np.sum(disc_masks, axis=(2, 3)).astype(float)

    # find the smallest p-value and determine if it's significant
    significant_cells = np.zeros((num_cells)).astype(bool)
    best_p = np.zeros((num_cells))
    p_value_correction_factor_per_pixel = (1.0 * num_y * num_x / pixels_per_mask_per_pixel)

    best_exclusion_region_list = []
    corrected_p_value_array_list = []
    for n in range(num_cells):

        # Sidak correction:
        p_value_corrected_per_pixel = 1-np.power((1-chi_square_grid[n, :,:]), p_value_correction_factor_per_pixel)
        corrected_p_value_array_list.append(p_value_corrected_per_pixel)

        y, x = np.unravel_index(p_value_corrected_per_pixel.argmin(), (num_y, num_x))

        # if more than one p-value that maxes out, use the median location
        if np.sum(p_value_corrected_per_pixel == 0.0) > 1:

            y, x = np.unravel_index(np.argwhere(p_value_corrected_per_pixel.flatten() == 0.0)[:, 0], (num_y, num_x))

            med_x = np.median(x)
            med_y = np.median(y)
            center_x = x[0]
            center_y = y[0]
            for i in range(len(x)):
                dx = x[i] - med_x
                dy = y[i] - med_y
                dc_x = center_x - med_x
                dc_y = center_y - med_y
                if np.sqrt(dx ** 2 + dy ** 2) < np.sqrt(dc_x ** 2 + dc_y ** 2):
                    center_x = x[i]
                    center_y = y[i]
            x = center_x
            y = center_y

        best_p[n] = p_value_corrected_per_pixel[y,x]
        if best_p[n] < alpha:
            significant_cells[n] = True
            best_exclusion_region_list.append(disc_masks[y, x, :,:].astype(np.bool))
        else:
            best_exclusion_region_list.append(np.zeros((disc_masks.shape[0], disc_masks.shape[1]), dtype=np.bool))

    return significant_cells, best_p, corrected_p_value_array_list, best_exclusion_region_list


def pvalue_to_NLL(p_values,
                  max_NLL=10.0):
    return np.where(p_values == 0.0, max_NLL, -np.log10(p_values))


def NLL_to_pvalue(NLLs,
                  log_base=10.0):
    return (log_base ** (-NLLs))


def get_events_per_pixel(responses_np,
                         trial_matrix):
    num_cells = np.shape(responses_np)[1]
    num_y = np.shape(trial_matrix)[0]
    num_x = np.shape(trial_matrix)[1]

    events_per_pixel = np.zeros((num_cells, num_y, num_x, 2))
    for y in range(num_y):
        for x in range(num_x):
            for on_off in range(2):
                frames = np.argwhere(trial_matrix[y, x, on_off, :])[:, 0]
                events_per_pixel[:, y, x, on_off] = np.sum(responses_np[frames, :], axis=0)

    return events_per_pixel


def smooth_STA(STA,
               gauss_std=0.75):

    deg_per_pnt = 64/STA.shape[0]
    STA_interpolated = interpolate_RF(STA, deg_per_pnt)
    STA_interpolated_smoothed = filt.gaussian_filter(STA_interpolated, gauss_std)
    STA_smoothed = deinterpolate_RF(STA_interpolated_smoothed, STA.shape[1], STA.shape[0], deg_per_pnt)

    return STA_smoothed


def interpolate_RF(rf_map,
                   deg_per_pnt):
    x_pnts = np.shape(rf_map)[1]
    y_pnts = np.shape(rf_map)[0]

    x_coor = np.arange(-(x_pnts - 1) * deg_per_pnt / 2, (x_pnts + 1) * deg_per_pnt / 2, deg_per_pnt)
    y_coor = np.arange(-(y_pnts - 1) * deg_per_pnt / 2, (y_pnts + 1) * deg_per_pnt / 2, deg_per_pnt)

    x_interpolated = np.arange(-(x_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (x_pnts / 2 - 1) * deg_per_pnt + 1, 1)
    y_interpolated = np.arange(-(y_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (y_pnts / 2 - 1) * deg_per_pnt + 1, 1)

    interpolated = si.interp2d(x_coor, y_coor, rf_map)
    interpolated = interpolated(x_interpolated, y_interpolated)

    return interpolated


def deinterpolate_RF(rf_map, x_pnts, y_pnts, deg_per_pnt):

    # x_pnts = 28
    # y_pnts = 16

    x_interpolated = np.arange(-(x_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (x_pnts / 2 - 1) * deg_per_pnt + 1, 1)
    y_interpolated = np.arange(-(y_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (y_pnts / 2 - 1) * deg_per_pnt + 1, 1)

    x_deinterpolate = np.arange(0, len(x_interpolated), deg_per_pnt)
    y_deinterpolate = np.arange(0, len(y_interpolated), deg_per_pnt)

    sampled_y = rf_map[y_deinterpolate, :]
    sampled_yx = sampled_y[:, x_deinterpolate]

    return sampled_yx


def chi_square_within_mask(exclusion_mask,
                           events_per_pixel,
                           trials_per_pixel):
    num_y = np.shape(exclusion_mask)[0]
    num_x = np.shape(exclusion_mask)[1]
    num_cells = np.shape(events_per_pixel)[0]

    # d.f. is number of pixels in mask minus one
    degrees_of_freedom = int(np.sum(exclusion_mask)) - 1

    # calculate expected number of events per pixel
    #   expected_by_pixel has shape (num_cells,num_y,num_x,2)
    num_trials_by_pixel = exclusion_mask * trials_per_pixel  # shape is (num_y,num_x,2)
    total_trials_in_mask = np.sum(num_trials_by_pixel).astype(float)
    events_by_pixel_in_mask = exclusion_mask.reshape(1, num_y, num_x, 2) * events_per_pixel
    total_events_in_mask = np.sum(events_by_pixel_in_mask, axis=(1, 2, 3)).astype(float)
    expected_per_trial = total_events_in_mask / total_trials_in_mask  # shape is (num_cells,)
    expected_by_pixel = num_trials_by_pixel.reshape(1, num_y, num_x, 2) * expected_per_trial.reshape(num_cells, 1, 1, 1)

    # calculate observed number of events per pixel
    #   observed_by_pixel has shape (num_cells,num_y,num_x,2)
    observed_by_pixel = (events_per_pixel * exclusion_mask.reshape(1, num_y, num_x, 2)).astype(float)

    # calculate test statistic given observed and expected
    residual_by_pixel = observed_by_pixel - expected_by_pixel
    chi = (residual_by_pixel ** 2) / expected_by_pixel
    chi_sum = np.nansum(chi, axis=(1, 2, 3))

    # get p-value given test statistic and degrees of freedom
    p_vals = 1.0 - stats.chi2.cdf(chi_sum, degrees_of_freedom)

    return p_vals, chi


def build_trial_matrix(LSN_template,
                       num_trials):
    num_y = np.shape(LSN_template)[1]
    num_x = np.shape(LSN_template)[2]
    on_off_luminance = [255, 0]

    trial_mat = np.zeros((num_y, num_x, 2, num_trials), dtype=bool)
    for y in range(num_y):
        for x in range(num_x):
            for on_off in range(2):
                frame = np.argwhere(LSN_template[:num_trials, y, x] == on_off_luminance[on_off])[:, 0]
                trial_mat[y, x, on_off, frame] = True

    return trial_mat


def get_disc_masks(LSN_template,
                   radius=3):
    on_luminance = 255
    off_luminance = 0
    num_y = np.shape(LSN_template)[1]
    num_x = np.shape(LSN_template)[2]

    # convert template to true on trials a pixel is not gray and false when gray
    LSN_binary = np.where(LSN_template == off_luminance, 1, LSN_template)
    LSN_binary = np.where(LSN_binary == on_luminance, 1, LSN_binary)
    LSN_binary = np.where(LSN_binary == 1, 1.0, 0.0)

    # get number of trials each pixel is not gray
    on_trials = np.sum(LSN_binary, axis=0).astype(float)  # shape is (num_y,num_x)

    masks = np.zeros((num_y, num_x, num_y, num_x))
    for y in range(num_y):
        for x in range(num_x):
            trials_not_gray = np.argwhere(LSN_binary[:, y, x] > 0)[:, 0]
            raw_mask = np.divide(np.sum(LSN_binary[trials_not_gray, :, :], axis=0), on_trials)

            center_y, center_x = np.unravel_index(raw_mask.argmax(), (num_y, num_x))

            # include center pixel in mask
            raw_mask[center_y, center_x] = 0.0

            x_max = center_x + radius + 1
            if x_max > num_x:
                x_max = num_x
            x_min = center_x - radius
            if x_min < 0:
                x_min = 0
            y_max = center_y + radius + 1
            if y_max > num_y:
                y_max = num_y
            y_min = center_y - radius
            if y_min < 0:
                y_min = 0

            # don't include far away pixels that just happen
            # to not have any trials in common with center pixel
            clean_mask = np.ones(np.shape(raw_mask))
            clean_mask[y_min:y_max, x_min:x_max] = raw_mask[y_min:y_max, x_min:x_max]

            masks[y, x, :, :] = clean_mask

    masks = np.where(masks > 0, 0.0, 1.0)

    return masks




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from allensdk.deprecated import deprecated, class_deprecated
import warnings


@pytest.fixture
def deprecated_method():

    @deprecated()
    def i_am_deprecated():
        pass

    return i_am_deprecated
    
    
@pytest.fixture
def deprecated_class():

    @class_deprecated('msg')
    class dep_cls(object):
        def __init__(self, a):
            self.a = a
        
    return dep_cls


def test_deprecated(deprecated_method):
    expected = "Function i_am_deprecated is deprecated. "

    with warnings.catch_warnings(record=True) as c:
        warnings.simplefilter('always')
        deprecated_method()

        print(expected)
        print(str(c[-1].message))

        assert expected == str(c[-1].message)
        
        
def test_deprecated_class(deprecated_class):
    expected = 'Class dep_cls is deprecated. msg'
    
    with warnings.catch_warnings(record=True) as c:
        warnings.simplefilter('always')
        deprecated_method()

        obj = deprecated_class(1)

        assert( expected == str(c[-1].message) )
        assert( obj.a == 1 )



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import simplejson as json
import math
import re
import logging

ju_logger = logging.getLogger(__name__)

try:
    import urllib.request as urllib_request
except ImportError:
    import urllib2 as urllib_request
try:
    from urllib.parse import urlparse
except ImportError:
    import urlparse


def read(file_name):
    """ Shortcut reading JSON from a file. """
    with open(file_name, 'rb') as f:
        json_string = f.read().decode('utf-8')
        if len(json_string)==0: # If empty file
            json_string='{}' # Create a string that will give an empty JSON object instead of an error
        json_obj = json.loads(json_string)

    return json_obj


def write(file_name, obj):
    """ Shortcut for writing JSON to a file.  This also takes care of serializing numpy and data types. """
    with open(file_name, 'wb') as f:
        try:
            f.write(write_string(obj))   # Python 2.7
        except TypeError:
            f.write(bytes(write_string(obj), 'utf-8'))  # Python 3


def write_string(obj):
    """ Shortcut for writing JSON to a string.  This also takes care of serializing numpy and data types. """
    return json.dumps(obj,
                      indent=2,
                      ignore_nan=True,
                      default=json_handler,
                      iterable_as_array=True)


def read_url(url, method='POST'):
    if method == 'GET':
        return read_url_get(url)
    elif method == 'POST':
        return read_url_post(url)
    else:
        raise Exception('Unknown request method: (%s)' % method)


def read_url_get(url):
    '''Transform a JSON contained in a file into an equivalent
    nested python dict.

    Parameters
    ----------
    url : string
        where to get the json.

    Returns
    -------
    dict
        Python version of the input

    Note: if the input is a bare array or literal, for example,
    the output will be of the corresponding type.
    '''
    response = urllib_request.urlopen(url)
    json_string = response.read().decode('utf-8')

    return json.loads(json_string)


def read_url_post(url):
    '''Transform a JSON contained in a file into an equivalent
    nested python dict.

    Parameters
    ----------
    url : string
        where to get the json.

    Returns
    -------
    dict
        Python version of the input

    Note: if the input is a bare array or literal, for example,
    the output will be of the corresponding type.
    '''
    urlp = urlparse.urlparse(url)
    main_url = urlparse.urlunsplit(
        (urlp.scheme, urlp.netloc, urlp.path, '', ''))
    data = json.dumps(dict(urlparse.parse_qsl(urlp.query)))

    handler = urllib_request.HTTPHandler()
    opener = urllib_request.build_opener(handler)

    request = urllib_request.Request(main_url, data)
    request.add_header("Content-Type", 'application/json')
    request.get_method = lambda: 'POST'

    try:
        response = opener.open(request)
    except Exception as e:
        response = e

    if response.code == 200:
        json_string = response.read()
    else:
        json_string = response.read()

    return json.loads(json_string)


def json_handler(obj):
    """ Used by write_json convert a few non-standard types to things that the json package can handle. """
    if hasattr(obj, 'to_dict'):
        return obj.to_dict()
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif (isinstance(obj, np.float64) or
          isinstance(obj, np.float32) or
          isinstance(obj, np.float16)):
        return float(obj)
    elif (isinstance(obj, np.int64) or
          isinstance(obj, np.int32) or
          isinstance(obj, np.int16) or
          isinstance(obj, np.int8) or
          isinstance(obj, np.uint64) or
          isinstance(obj, np.uint32) or
          isinstance(obj, np.uint16) or
          isinstance(obj, np.uint8)):
        return long(obj)
    elif (isinstance(obj, np.bool) or
          isinstance(obj, np.bool_)):
        return bool(obj)
    elif hasattr(obj, 'isoformat'):
        return obj.isoformat()
    else:
        raise TypeError(
            'Object of type %s with value of %s is not JSON serializable' %
            (type(obj), repr(obj)))


class JsonComments(object):
    _oneline_comment = re.compile(r"\/\/.*$",
                                  re.MULTILINE)
    _multiline_comment_start = re.compile(r"\/\*",
                                          re.MULTILINE |
                                          re.DOTALL)
    _multiline_comment_end = re.compile(r"\*\/",
                                        re.MULTILINE |
                                        re.DOTALL)
    _blank_line = re.compile(r"\n?^\s*$", re.MULTILINE)
    _carriage_return = re.compile(r"\r$", re.MULTILINE)

    @classmethod
    def read_string(cls, json_string):
        json_string_no_comments = cls.remove_comments(json_string)
        return json.loads(json_string_no_comments)

    @classmethod
    def read_file(cls, file_name):
        try:
            with open(file_name) as f:
                json_string = f.read()
                json_object = cls.read_string(json_string)

                return json_object
        except ValueError:
            ju_logger.error(
                "Could not load json object from file: %s" % (file_name))
            raise

    @classmethod
    def remove_comments(cls, json_string):
        '''Strip single and multiline javascript-style comments.

        Parameters
        ----------
        json : string
            Json string with javascript-style comments.

        Returns
        -------
        string
            Copy of the input with comments removed.

        Note: A JSON decoder MAY accept and ignore comments.
        '''
        json_string = JsonComments._oneline_comment.sub('', json_string)
        json_string = JsonComments._carriage_return.sub('', json_string)
        json_string = JsonComments.remove_multiline_comments(json_string)
        json_string = JsonComments._blank_line.sub('', json_string)

        return json_string

    @classmethod
    def remove_multiline_comments(cls, json_string):
        '''Rebuild input without substrings matching /*...*/.

        Parameters
        ----------
        json_string : string
            may or may not contain multiline comments.

        Returns
        -------
        string
            Copy of the input without the comments.
        '''
        new_json = []
        start_iter = JsonComments._multiline_comment_start.finditer(
            json_string)
        json_slice_start = 0

        for comment_start in start_iter:
            json_slice_end = comment_start.start()
            new_json.append(json_string[json_slice_start:json_slice_end])
            search_start = comment_start.end()
            comment_end = JsonComments._multiline_comment_end.search(json_string[
                                                                     search_start:])
            if comment_end is None:
                break
            else:
                json_slice_start = search_start + comment_end.end()
        new_json.append(json_string[json_slice_start:])

        return ''.join(new_json)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2014-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
import re
import logging
import errno
import pandas as pd

class ManifestVersionError(Exception): 
    def __init__(self, message, version, found_version):
        super(ManifestVersionError, self).__init__(message)
        self.found_version = found_version
        self.version = version
        

class Manifest(object):
    """Manages the location of external files
     referenced in an Allen SDK configuration """

    DIR = 'dir'
    FILE = 'file'
    DIRNAME = 'dir_name'
    VERSION = 'manifest_version'

    log = logging.getLogger(__name__)

    def __init__(self, config=None, relative_base_dir='.', version=None):
        self.path_info = {}
        self.relative_base_dir = relative_base_dir

        if config is not None:
            self.load_config(config, version=version)

    def load_config(self, config, version=None):
        ''' Load paths into the manifest from an Allen SDK config section.

        Parameters
        ----------
        config : Config
            Manifest section of an Allen SDK config.
        '''
        found_version = None
        for path_info in config:
            path_type = path_info['type']
            path_format = None
            if 'format' in path_info:
                path_format = path_info['format']

            if path_type == 'file':
                try:
                    parent_key = path_info['parent_key']
                except:
                    parent_key = None

                self.add_file(path_info['key'],
                              path_info['spec'],
                              parent_key,
                              path_format)
            elif path_type == 'dir':
                try:
                    parent_key = path_info['parent_key']
                except:
                    parent_key = None

                spec = path_info['spec']
                absolute = False
                if spec[0] == '/':
                    absolute = True
                self.add_path(path_info['key'],
                              path_info['spec'],
                              path_type,
                              absolute,
                              path_format,
                              parent_key)

            elif path_type == self.VERSION:
                found_version = path_info['value']
            else:
                Manifest.log.warning("Unknown path type in manifest: %s" %
                                     (path_type))


        if found_version != version:
            raise ManifestVersionError("", version, found_version)
        self.version = version

    def add_path(self, key, path, path_type=DIR,
                 absolute=True, path_format=None, parent_key=None):
        '''Insert a new entry.

        Parameters
        ----------
        key : string
            Identifier for referencing the entry.
        path : string
            Specification for a path using %s, %d style substitution.
        path_type : string enumeration
            'dir' (default) or 'file'
        absolute : boolean
            Is the spec relative to the process current directory.
        path_format : string, optional
            Indicate a known file type for further parsing.
        parent_key : string
            Refer to another entry.
        '''
        if parent_key:
            path_args = []

            try:
                parent_path = self.path_info[parent_key]['spec']
                path_args.append(parent_path)
            except:
                Manifest.log.error(
                    "cannot resolve directory key %s" % (parent_key))
                raise
            path_args.extend(path.split('/'))
            path = os.path.join(*path_args)

        # TODO: relative paths need to be considered better
        if absolute is True:
            path = os.path.abspath(path)
        else:
            path = os.path.abspath(os.path.join(self.relative_base_dir, path))

        if path_type == Manifest.DIRNAME:
            path = os.path.dirname(path)

        self.path_info[key] = {'type': path_type,
                               'spec': path}

        if path_type == Manifest.FILE and path_format is not None:
            self.path_info[key]['format'] = path_format

    def add_paths(self, path_info):
        ''' add information about paths stored in the manifest.

        Parameters
            path_info : dict
                Information about the new paths
        '''
        for path_key, path_data in path_info.items():
            path_format = None

            if 'format' in path_data:
                path_format = path_data['format']

            Manifest.log.info("Adding path.  type: %s, format: %s, spec: %s" %
                              (path_data['type'],
                               path_data['spec'],
                               path_format))
            entry = {'type': path_data['type'],
                     'spec': path_data['spec']
                     }
            if path_format is not None:
                entry['format'] = path_format

            self.path_info[path_key] = entry

    def add_file(self,
                 file_key,
                 file_name,
                 dir_key=None,
                 path_format=None):
        '''Insert a new file entry.

        Parameters
        ----------
        file_key : string
            Reference to the entry.
        file_name : string
            Subtitutions of the %s, %d style allowed.
        dir_key : string
            Reference to the parent directory entry.
        path_format : string, optional
            File type for further parsing.
        '''
        path_args = []

        if dir_key:
            try:
                dir_path = self.path_info[dir_key]['spec']
                path_args.append(dir_path)
            except:
                Manifest.log.error(
                    "cannot resolve directory key %s" % (dir_key))
                raise
        elif not file_name.startswith('/'):
            path_args.append(os.curdir)
        else:
            path_args.append(os.path.sep)

        path_args.extend(file_name.split('/'))
        file_path = os.path.join(*path_args)

        self.path_info[file_key] = {'type': Manifest.FILE,
                                    'spec': file_path}

        if path_format:
            self.path_info[file_key]['format'] = path_format

    def get_path(self, path_key, *args):
        '''Retrieve an entry with substitutions.

        Parameters
        ----------
        path_key : string
            Refer to the entry to retrieve.
        args : any types, optional
           arguments to be substituted into the path spec for %s, %d, etc.

        Returns
        -------
        string
            Path with parent structure and substitutions applied.
        '''
        path_spec = self.path_info[path_key]['spec']

        if args is not None and len(args) != 0:
            path = path_spec % args
        else:
            path = path_spec

        return path

    def get_format(self, path_key):
        '''Retrieve the type of a path entry.

        Parameters
        ----------
        path_key : string
            reference to the entry

        Returns
        -------
        string
            File type.
        '''
        path_entry = self.path_info[path_key]
        path_format = None

        if 'format' in path_entry:
            path_format = path_entry['format']

        return path_format

    @classmethod
    def safe_make_parent_dirs(cls, file_name):
        ''' Create a parent directories for file.

        Parameters
        ----------
        file_name : string
        '''

        dirname = os.path.dirname(file_name)

        # do nothing if there are no parent directories
        if not dirname:
            return

        Manifest.safe_mkdir(dirname)

    @classmethod
    def safe_mkdir(cls, directory):
        '''Create path if not already there.

        Parameters
        ----------
        directory : string
            create it if it doesn't exist
        '''
        

        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise

    def create_dir(self, path_key):
        '''Make a directory for an entry.

        Parameters
        ----------
        path_key : string
            Reference to the entry.
        '''
        dir_path = self.get_path(path_key)
        Manifest.safe_mkdir(dir_path)

    def check_dir(self, path_key, do_exit=False):
        '''Verify a directories existence or optionally exit.

        Parameters
        ----------
        path_key : string
            Reference to the entry.
        do_exit : boolean
            What to do if the directory is not present.
        '''
        dir_path = self.get_path(path_key)

        if not os.path.exists(dir_path):
            Manifest.log.fatal('Directory %s does not exist; exiting.' %
                               (dir_path))
            if do_exit is True:
                quit()

    def resolve_paths(self, description_dict, suffix='_key'):
        '''Walk input items and expand those that refer to a manifest entry.

        Parameters
        ----------
        description_dict : dict
            Any entries with key names ending in suffix will be expanded.
        suffix : string
            Indicates the entries to be expanded.
        '''
        key_pattern = re.compile('(.*)%s$' % (suffix))

        for description_key, manifest_key in description_dict.items():
            m = key_pattern.match(description_key)
            if m:
                real_key = m.group(1)  # i.e. job_dir_key -> job_dir
                filename = self.get_path(manifest_key)
                description_dict[real_key] = filename
                del description_dict[description_key]

    def as_dataframe(self):
        return pd.DataFrame.from_dict(self.path_info,
                                      orient='index')



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import patch, mock_open
from allensdk.model.biophys_sim.config import Config
try:
    import __builtin__ as builtins  # @UnresolvedImport
except:
    import builtins  # @UnresolvedImport


@pytest.fixture
def simple_config():
    manifest = '''{
        "manifest": [
            { "type": "dir",
              "spec": "MOCK_DOT",
              "key": "BASEDIR"
            }],
        "biophys":
            [{ "hoc": [ "stdgui.hoc"] }]
    }'''

    with patch(builtins.__name__ + ".open",
               mock_open(read_data=manifest)):
        config = Config().load('config.json', False)

    return config


def testAccessHocFilesInData(simple_config):
    assert simple_config.data['biophys'][0]['hoc'][0] == 'stdgui.hoc'


def testManifestIsNotInData(simple_config):
    assert 'manifest' not in simple_config.data


def testManifestInReservedData(simple_config):
    assert 'manifest' in simple_config.reserved_data[0]



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import argparse
import allensdk.core.swc as swc
try:
    xrange
except:
    from past.builtins import xrange


def validate_swc(swc_file):
    """
    To be compatible with NEURON, SWC files must have the following properties:
        1) a single root node with parent ID '-1'
        2) sequentially increasing ID numbers
        3) immediate children of the soma cannot branch
    """
    soma_id = swc.Morphology.SOMA
    morphology = swc.read_swc(swc_file)
    # verify that there is a single root node
    num_soma_nodes = sum([(int(c['type']) == soma_id)
                          for c in morphology.compartment_list])
    if num_soma_nodes != 1:
        raise Exception(
            "SWC must have single soma compartment.  Found: %d" % num_soma_nodes)
    # sanity check
    root = morphology.root
    if root is None:
        raise Exception("Morphology has no root node")
    # verify that children of the root have max one child
    for root_child_id in root['children']:
        root_child = morphology.compartment_index[root_child_id]
        num_grand_children = len(root_child['children'])
        if num_grand_children > 1:
            raise Exception("Child of root (%s) has more than one child (%d)" % (
                root_child_id, num_grand_children))
    # get a list of all of the ids, make sure they are unique while we're at it
    all_ids = set()
    for compartment in morphology.compartment_list:
        iid = int(compartment["id"])
        if iid in all_ids:
            raise Exception("Compartment ID %s is not unique." %
                            compartment["id"])
        pid = int(compartment["parent"])
        if iid < pid:
            raise Exception(
                "Compartment (%d) has a smaller ID that its parent (%d)" % (iid, pid))
        all_ids.add(iid)

    # sort the ids and make sure there are no gaps
    sorted_ids = sorted(all_ids)
    for i in xrange(1, len(sorted_ids)):
        if sorted_ids[i] - sorted_ids[i - 1] != 1:
            raise Exception("Compartment IDs are not sequential")
    return True


def main():
    try:
        parser = argparse.ArgumentParser(
            "validate an SWC file for use with NEURON")
        parser.add_argument('swc_file')
        args = parser.parse_args()
        validate_swc(args.swc_file)
    except Exception as e:
        print(str(e))
        exit(1)
if __name__ == "__main__":
    main()



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import functools
import operator as op
from collections import defaultdict
from six import iteritems

class SimpleTree( object ):
    def __init__(self, nodes, 
                 node_id_cb, 
                 parent_id_cb):
        '''A tree structure
        
        Parameters
        ----------
        nodes : list of dict
            Each dict is a node in the tree. The keys of the dict name the 
            properties of the node and should be consistent across nodes.
        node_id_cb : function | node dict -> node id
            Calling node_id_cb on a node dictionary ought to produce a unique 
            identifier for that node (we call this the node's id). The type 
            of the node id is up to you, but ought to be consistent across 
            nodes and must be hashable.
        parent_id_cb : function | node_dict => parent node's id
            As node_id_cb, but returns the id of the node's parent.
            
        Notes
        -----
        It is easy to pass a pandas DataFrame as the nodes. Just use the 
        to_dict method of the dataframe like so:
            list_of_dict = your_dataframe.to_dict('record')
            your_tree = SimpleTree(list_of_dict, ...)
        Converting a list of dictionaries to a pandas DataFrame is also very 
        easy. The DataFrame constructor does it for you:
            your_dataframe = pandas.DataFrame(list_of_dict)
             
        '''

        self._nodes = defaultdict(lambda *x: None, { node_id_cb(n):n for n in nodes })
        self._parent_ids = defaultdict(lambda *x: None, { nid:parent_id_cb(n) for nid,n in iteritems(self._nodes) })
        self._child_ids = defaultdict(lambda *x: None, { nid:[] for nid in self._nodes })

        for nid in self._parent_ids:
            pid = self._parent_ids[nid]
            if pid is not None:
                self._child_ids[pid].append(nid)


    def filter_nodes(self, criterion):
        '''Obtain a list of nodes filtered by some criterion
        
        Parameters
        ----------
        criterion : function | node dict => bool
            Only nodes for which criterion returns true will be returned.
            
        Returns
        -------
        list of dict :
            Items are node dictionaries that passed the filter.
        
        '''
    
        return list(filter(criterion, self._nodes.values()))
        
        
    def value_map(self, from_fn, to_fn):
        '''Obtain a look-up table relating a pair of node properties across 
        nodes
        
        Parameters
        ----------
        from_fn : function | node dict => hashable value
            The keys of the output dictionary will be obtained by calling 
            from_fn on each node.
        to_fn : function | node_dict => value
            The values of the output function will be obtained by calling 
            to_fn on each node.
            
        Returns
        -------
        dict :
            Maps the node property defined by from_fn to the node property 
            defined by to_fn across nodes.
            
        Notes
        -----
        The resulting map is not necessarily 1-to-1! 
        
        '''
    
        return {from_fn(v): to_fn(v) for v in self._nodes.values()}


    def node_ids(self):
        '''Obtain the node ids of each node in the tree
        
        Returns
        -------
        list :
            elements are node ids 
        
        '''
    
        return self._nodes.keys()


    def parent_id(self, node_ids):
        '''Obtain the ids of one or more nodes' parents
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose parents you wish to find.
        
        Returns
        -------
        list of hashable : 
            Items are ids of input nodes' parents in order.
        
        '''
        
        return [ self._parent_ids[nid] for nid in node_ids ]

   
    def child_ids(self, node_ids):
        '''Obtain the ids of one or more nodes' children
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose children you wish to find.
            
        Returns
        -------
        list of list of hashable : 
            Items are lists of input nodes' children's ids.
            
        '''
    
        return [ self._child_ids[nid] for nid in node_ids ]


    def ancestor_ids(self, node_ids):
        '''Obtain the ids of one or more nodes' ancestors
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose ancestors you wish to find.
        
        Returns
        -------
        list of list of hashable : 
            Items are lists of input nodes' ancestors' ids.
        
        Notes
        -----
        Given the tree:
        A -> B -> C
         `-> D
          
        The ancestors of C are [C, B, A]. The ancestors of A are [A]. The 
        ancestors of D are [D, A]
        
        '''
    
        out = []
        for nid in node_ids:
        
            current = [nid]
            while current[-1] is not None:
                current.extend(self.parent_id([current[-1]]))
            out.append(current[:-1])
                
        return out
            
    
    def descendant_ids(self, node_ids):
        '''Obtain the ids of one or more nodes' descendants
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose descendants you wish to find.
        
        Returns
        -------
        list of list of hashable : 
            Items are lists of input nodes' descendants' ids.
        
        Notes
        -----
        Given the tree:
        A -> B -> C
         `-> D
          
        The descendants of A are [B, C, D]. The descendants of C are [].
        
        '''
    
        out = []
        for ii, nid in enumerate(node_ids):
        
            current = [nid]
            children = self.child_ids([nid])[0]
            
            if children:
                current.extend(functools.reduce(op.add, map(list, 
                               self.descendant_ids(children))))
                               
            out.append(current)
        return out
            

    def node(self, node_ids=None):
        '''Get one or more nodes' full dictionaries from their ids.
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes to be returned. Default is all.
            
        Returns
        -------
        list of dict : 
            Items are nodes corresponding to argued ids.
        '''
    
        if node_ids is None:
            node_ids = self.node_ids()
    
        return [ self._nodes[nid] for nid in node_ids ]


    def parent(self, node_ids):
        '''Get one or mode nodes' parent nodes
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose parents will be found.
            
        Returns
        -------
        list of dict : 
            Items are parents of nodes corresponding to argued ids.
        
        '''
        
        return self.node([self._parent_ids[nid] for nid in node_ids])


    def children(self, node_ids):
        '''Get one or mode nodes' child nodes
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose children will be found.
            
        Returns
        -------
        list of list of dict : 
            Items are lists of child nodes corresponding to argued ids.
        
        '''
    
        return list(map(self.node, self.child_ids(node_ids)))


    def descendants(self, node_ids):
        '''Get one or mode nodes' descendant nodes
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose descendants will be found.
            
        Returns
        -------
        list of list of dict : 
            Items are lists of descendant nodes corresponding to argued ids.
        
        '''
        
        return list(map(self.node, self.descendant_ids(node_ids)))

    
    def ancestors(self, node_ids):
        '''Get one or mode nodes' ancestor nodes
        
        Parameters
        ----------
        node_ids : list of hashable
            Items are ids of nodes whose ancestors will be found.
            
        Returns
        -------
        list of list of dict : 
            Items are lists of ancestor nodes corresponding to argued ids.
        
        '''
    
        return list(map(self.node, self.ancestor_ids(node_ids)))
    



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import requests
from contextlib import closing
import allensdk.core.json_utilities as json_utilities
import pandas as pd
import logging
import os
import errno
import warnings


class Api(object):
    _log = logging.getLogger(__name__)
    default_api_url = 'http://api.brain-map.org'
    download_url = 'http://download.alleninstitute.org'

    def __init__(self, api_base_url_string=None):
        if api_base_url_string is None:
            api_base_url_string = Api.default_api_url

        self.set_api_urls(api_base_url_string)
        self.default_working_directory = os.getcwd()

    def set_api_urls(self, api_base_url_string):
        '''Set the internal RMA and well known file download endpoint urls
        based on a api server endpoint.

        Parameters
        ----------
        api_base_url_string : string
            url of the api to point to
        '''
        self.api_url = api_base_url_string

        # http://help.brain-map.org/display/api/Downloading+a+WellKnownFile
        self.well_known_file_endpoint = api_base_url_string + \
            '/api/v2/well_known_file_download'

        # http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data
        self.grid_data_endpoint = api_base_url_string + '/grid_data'

        # http://help.brain-map.org/display/api/Downloading+and+Displaying+SVG
        self.svg_endpoint = api_base_url_string + '/api/v2/svg'
        self.svg_download_endpoint = api_base_url_string + '/api/v2/svg_download'

        # http://help.brain-map.org/display/api/Downloading+an+Ontology%27s+Structure+Graph
        self.structure_graph_endpoint = api_base_url_string + \
            '/api/v2/structure_graph_download'

        # http://help.brain-map.org/display/api/Searching+a+Specimen+or+Structure+Tree
        self.tree_search_endpoint = api_base_url_string + '/api/v2/tree_search'

        # http://help.brain-map.org/display/api/Searching+Annotated+SectionDataSets
        self.annotated_section_data_sets_endpoint = api_base_url_string + \
            '/api/v2/annotated_section_data_sets'
        self.compound_annotated_section_data_sets_endpoint = api_base_url_string + \
            '/api/v2/compound_annotated_section_data_sets'

        # http://help.brain-map.org/display/api/Image-to-Image+Synchronization#Image-to-ImageSynchronization-ImagetoImage
        self.image_to_atlas_endpoint = api_base_url_string + '/api/v2/image_to_atlas'
        self.image_to_image_endpoint = api_base_url_string + '/api/v2/image_to_image'
        self.image_to_image_2d_endpoint = api_base_url_string + '/api/v2/image_to_image_2d'
        self.reference_to_image_endpoint = api_base_url_string + '/api/v2/reference_to_image'
        self.image_to_reference_endpoint = api_base_url_string + '/api/v2/image_to_reference'
        self.structure_to_image_endpoint = api_base_url_string + '/api/v2/structure_to_image'

        # http://help.brain-map.org/display/mouseconnectivity/API
        self.section_image_download_endpoint = api_base_url_string + \
            '/api/v2/section_image_download'
        self.atlas_image_download_endpoint = api_base_url_string + \
            '/api/v2/atlas_image_download'
        self.projection_image_download_endpoint = api_base_url_string + \
            '/api/v2/projection_image_download'
        self.image_download_endpoint = api_base_url_string + \
            '/api/v2/image_download'
        self.informatics_archive_endpoint = Api.download_url + '/informatics-archive'

        self.rma_endpoint = api_base_url_string + '/api/v2/data'

    def set_default_working_directory(self, working_directory):
        '''Set the working directory where files will be saved.

        Parameters
        ----------
        working_directory : string
             the absolute path string of the working directory.
        '''
        self.default_working_directory = working_directory

    def read_data(self, parsed_json):
        '''Return the message data from the parsed query.

        Parameters
        ----------
        parsed_json : dict
            A python structure corresponding to the JSON data returned from the API.

        Notes
        -----
        See `API Response Formats - Response Envelope <http://help.brain-map.org/display/api/API+Response+Formats#APIResponseFormats-ResponseEnvelope>`_
        for additional documentation.
        '''
        return parsed_json['msg']

    def json_msg_query(self, url, dataframe=False):
        ''' Common case where the url is fully constructed
            and the response data is stored in the 'msg' field.

        Parameters
        ----------
        url : string
            Where to get the data in json form
        dataframe : boolean
            True converts to a pandas dataframe, False (default) doesn't

        Returns
        -------
        dict or DataFrame
            returned data; type depends on dataframe option
        '''

        data = self.do_query(lambda *a, **k: url,
                             self.read_data)

        if dataframe is True:
            warnings.warn("dataframe argument is deprecated", DeprecationWarning)
            data = pd.DataFrame(data)

        return data

    def do_query(self, url_builder_fn, json_traversal_fn, *args, **kwargs):
        '''Bundle an query url construction function
        with a corresponding response json traversal function.

        Parameters
        ----------
        url_builder_fn : function
            A function that takes parameters and returns an rma url.
        json_traversal_fn : function
            A function that takes a json-parsed python data structure and returns data from it.
        post : boolean, optional kwarg
            True does an HTTP POST, False (default) does a GET
        args : arguments
            Arguments to be passed to the url builder function.
        kwargs : keyword arguments
            Keyword arguments to be passed to the rma builder function.

        Returns
        -------
        any type
            The data extracted from the json response.

        Examples
        --------
        `A simple Api subclass example
        <data_api_client.html#creating-new-api-query-classes>`_.
        '''
        api_url = url_builder_fn(*args, **kwargs)

        post = kwargs.get('post', False)

        json_parsed_data = self.retrieve_parsed_json_over_http(api_url, post)

        return json_traversal_fn(json_parsed_data)

    def do_rma_query(self, rma_builder_fn, json_traversal_fn, *args, **kwargs):
        '''Bundle an RMA query url construction function
        with a corresponding response json traversal function.

        ..note:: Deprecated in AllenSDK 0.9.2
            `do_rma_query` will be removed in AllenSDK 1.0, it is replaced by
            `do_query` because the latter is more general.

        Parameters
        ----------
        rma_builder_fn : function
            A function that takes parameters and returns an rma url.
        json_traversal_fn : function
            A function that takes a json-parsed python data structure and returns data from it.
        args : arguments
            Arguments to be passed to the rma builder function.
        kwargs : keyword arguments
            Keyword arguments to be passed to the rma builder function.

        Returns
        -------
        any type
            The data extracted from the json response.

        Examples
        --------
        `A simple Api subclass example
        <data_api_client.html#creating-new-api-query-classes>`_.
        '''
        return self.do_query(rma_builder_fn, json_traversal_fn, *args, **kwargs)

    def load_api_schema(self):
        '''Download the RMA schema from the current RMA endpoint

        Returns
        -------
        dict
            the parsed json schema message

        Notes
        -----
        This information and other
        `Allen Brain Atlas Data Portal Data Model <http://help.brain-map.org/display/api/Data+Model>`_
        documentation is also available as a
        `Class Hierarchy <http://api.brain-map.org/class_hierarchy>`_
        and `Class List <http://api.brain-map.org/class_hierarchy>`_.

        '''
        schema_url = self.rma_endpoint + '/enumerate.json'
        json_parsed_schema_data = self.retrieve_parsed_json_over_http(
            schema_url)

        return json_parsed_schema_data

    def construct_well_known_file_download_url(self, well_known_file_id):
        '''Join data api endpoint and id.

        Parameters
        ----------
        well_known_file_id : integer or string representing an integer
            well known file id

        Returns
        -------
        string
            the well-known-file download url for the current api api server

        See Also
        --------
        retrieve_file_over_http: Can be used to retrieve the file from the url.
        '''
        return self.well_known_file_endpoint + '/' + str(well_known_file_id)

    def cleanup_truncated_file(self, file_path):
        '''Helper for removing files.

        Parameters
        ----------
        file_path : string
            Absolute path including the file name to remove.'''
        try:
            os.remove(file_path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    def retrieve_file_over_http(self, url, file_path):
        '''Get a file from the data api and save it.

        Parameters
        ----------
        url : string
            Url[1]_ from which to get the file.
        file_path : string
            Absolute path including the file name to save.

        See Also
        --------
        construct_well_known_file_download_url: Can be used to construct the url.

        References
        ----------
        .. [1] Allen Brain Atlas Data Portal: `Downloading a WellKnownFile <http://help.brain-map.org/display/api/Downloading+a+WellKnownFile>`_.
        '''

        self._log.info("Downloading URL: %s", url)
        
        from requests_toolbelt import exceptions
        from requests_toolbelt.downloadutils import stream

        try:
            with closing(requests.get(url,
                                      stream=True,
                                      timeout=(9.05,31.1))) as response:
                response.raise_for_status()
                with open(file_path, 'wb') as f:
                    stream.stream_response_to_file(response, path=f)
        except exceptions.StreamingError as e:
            self._log.error("Couldn't retrieve file %s from %s (streaming)." % (file_path,url))
            self.cleanup_truncated_file(file_path)
            raise
        except requests.exceptions.ConnectionError as e:
            self._log.error("Couldn't retrieve file %s from %s (connection)." % (file_path,url))
            self.cleanup_truncated_file(file_path)
            raise
        except requests.exceptions.ReadTimeout as e:
            self._log.error("Couldn't retrieve file %s from %s (timeout)." % (file_path,url))
            self.cleanup_truncated_file(file_path)
            raise
        except requests.exceptions.RequestException as e:
            self._log.error("Couldn't retrieve file %s from %s (request)." % (file_path,url))
            self.cleanup_truncated_file(file_path)
            raise
        except Exception as e:
            self._log.error("Couldn't retrieve file %s from %s" % (file_path, url))
            self.cleanup_truncated_file(file_path)
            raise


    def retrieve_parsed_json_over_http(self, url, post=False):
        '''Get the document and put it in a Python data structure

        Parameters
        ----------
        url : string
            Full API query url.
        post : boolean
            True does an HTTP POST, False (default) encodes the URL and does a GET

        Returns
        -------
        dict
            Result document as parsed by the JSON library.
        '''
        self._log.info("Downloading URL: %s", url)
        
        if post is False:
            data = json_utilities.read_url_get(
                requests.utils.quote(url,
                                     ';/?:@&=+$,'))
        else:
            data = json_utilities.read_url_post(url)

        return data

    def retrieve_xml_over_http(self, url):
        '''Get the document and put it in a Python data structure

        Parameters
        ----------
        url : string
            Full API query url.

        Returns
        -------
        string
            Unparsed xml string.
        '''
        self._log.info("Downloading URL: %s", url)
                
        response = requests.get(url)

        return response.content



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import logging
import os
import argparse
import matplotlib.pyplot as plt
import h5py
import numpy as np

from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet


def movingmode_fast(x, kernelsize, y):
    """ Compute the windowed mode of an array.  A running mode is initialized
    with a histogram of values over the initial kernelsize/2 values.  The mode
    is then updated as the kernel moves by adding and subtracting values from
    the histogram.

    Parameters
    ----------
    x: np.ndarray
        Array to be analyzed
    kernelsize: int
        Size of the moving window
    y: np.ndarray
        Output array to store the results
    """

    # offset so that the trace is non-negative
    minval = min(x.min(), 0)
    if minval < 0:
        x = x - minval

    maxval = x.max()

    # compute a histogram of a half kernel
    halfsize = int(kernelsize / 2)
    histo = np.bincount(np.rint(x[:halfsize]).astype(
        np.uint32), minlength=int(maxval + 2))

    # find the mode of the first half kernel
    mode = np.argmax(histo)

    # here initial mode is available
    for m in range(0, halfsize):
        q = int(round(x[halfsize + m]))

        histo[q] += 1

        if histo[q] > histo[mode]:
            mode = q

        y[m] = mode

    for m in range(halfsize, x.shape[0] - halfsize):
        p = int(round(x[m - halfsize]))
        histo[p] -= 1

        # need to find possibly new mode value
        if p == mode:
            mode = np.argmax(histo)

        q = int(round(x[m + halfsize]))

        histo[q] += 1

        if histo[q] > histo[mode]:
            mode = q

        y[m] = mode

    for m in range(x.shape[0] - halfsize, x.shape[0]):
        p = int(round(x[m - halfsize]))
        histo[p] -= 1

        # need to find possibly new mode value
        if p == mode:
            mode = np.argmax(histo)

        y[m] = mode

    # undo the offset
    if minval < 0:
        y += minval

    return 0


def movingaverage(x, kernelsize, y):
    """ Compute the windowed average of an array.

    Parameters
    ----------
    x: np.ndarray
        Array to be analyzed
    kernelsize: int
        Size of the moving window
    y: np.ndarray
        Output array to store the results
    """

    halfsize = kernelsize / 2
    sumkernel = np.sum(x[0:halfsize])
    for m in range(0, halfsize):
        sumkernel = sumkernel + x[m + halfsize]
        y[m] = sumkernel / (halfsize + m)

    sumkernel = np.sum(x[0:kernelsize])
    for m in range(halfsize, x.shape[0] - halfsize):
        sumkernel = sumkernel - x[m - halfsize] + x[m + halfsize]
        y[m] = sumkernel / kernelsize

    for m in range(x.shape[0] - halfsize, x.shape[0]):
        sumkernel = sumkernel - x[m - halfsize]
        y[m] = sumkernel / (halfsize - 1 + (x.shape[0] - m))

    return 0


def plot_onetrace(dff, fc):
    """ Debug plotting function """
    qs = np.rint(np.linspace(0, len(dff), 5)).astype(int)

    dff_max = dff.max()
    dff_min = dff.min()
    fc_max = fc.max()
    fc_min = fc.min()

    for qi in range(len(qs) - 1):
        r = qs[qi], qs[qi + 1]

        frames = np.arange(r[0], r[1])
        ax = plt.subplot(len(qs), 1, qi + 1)
        ax.plot(frames, dff[r[0]:r[1]], 'g')
        ax.set_ylim(dff_min, dff_max)
        ax.set_xlim(r[0], r[1])
        ax.set_xlabel('frames', fontsize=18)
        ax.set_ylabel('DF/F', fontsize=18, color='g')

        ax = ax.twinx()
        ax.plot(frames, fc[r[0]:r[1]], 'b')
        ax.set_ylim(fc_min, fc_max)
        ax.set_xlim(r[0], r[1])
        ax.set_ylabel('FC', fontsize=18, color='b')

    return 0


def compute_dff(traces, save_plot_dir=None, mode_kernelsize=5400, mean_kernelsize=3000):
    """ Compute dF/F of a set of traces using a low-pass windowed-mode operator.
    The operation is basically:

        T_mm = windowed_mean(windowed_mode(T))

        T_dff = (T - T_mm) / T_mm

    Parameters
    ----------
    traces: np.ndarray
       2D array of traces to be analyzed

    Returns
    -------
    np.ndarray with the same shape as the input array.
    """

    if mode_kernelsize >= traces.shape[1]:
        raise Exception("Cannot compute dF/F: mode filter size (%d) longer than trace (%d)" %
                        (mode_kernelsize, traces.shape[1]))
    if mean_kernelsize >= traces.shape[1]:
        raise Exception("Cannot compute dF/F: mean filter size (%d) longer than trace (%d)" %
                        (mean_kernelsize, traces.shape[1]))

    if save_plot_dir is not None and not os.path.exists(save_plot_dir):
        os.makedirs(save_plot_dir)

    logging.debug("trace matrix shape: %d %d" %
                  (traces.shape[0], traces.shape[1]))

    modeline = np.zeros(traces.shape[1])
    modelineLP = np.zeros(traces.shape[1])
    dff = np.zeros((traces.shape[0], traces.shape[1]))

    logging.debug("computing df/f")

    for n in range(0, traces.shape[0]):
        if np.any(np.isnan(traces[n])):
            logging.warning("trace for roi %d contains NaNs, setting to NaN", n)
            dff[n,:] = np.nan
            continue

        movingmode_fast(traces[n, :], mode_kernelsize, modeline[:])
        movingaverage(modeline[:], mean_kernelsize, modelineLP[:])
        dff[n, :] = (traces[n, :] - modelineLP[:]) / modelineLP[:]

        logging.debug("finished trace %d/%d" % (n + 1, traces.shape[0]))

        if save_plot_dir:
            fig = plt.figure(figsize=(150, 40))
            plot_onetrace(dff[n, :], traces[n, :])

            plt.title('ROI ' + str(n) + ' ', fontsize=18)
            fig.savefig(os.path.join(save_plot_dir, 'dff_%d.png' %
                                     n), orientation='landscape')
            plt.close(fig)

    return dff


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_h5")
    parser.add_argument("output_h5")
    parser.add_argument("--plot_dir")
    parser.add_argument("--log_level", default=logging.INFO)

    args = parser.parse_args()

    logging.getLogger().setLevel(args.log_level)

    # read from "data"
    if args.input_h5.endswith("nwb"):
        timestamps, traces = BrainObservatoryNwbDataSet(
            args.input_h5).get_corrected_fluorescence_traces()
    else:
        input_h5 = h5py.File(args.input_h5, "r")
        traces = input_h5["data"].value
        input_h5.close()

    dff = compute_dff(traces, args.plot_dir)

    # write to "data"
    output_h5 = h5py.File(args.output_h5, "w")
    output_h5["data"] = dff
    output_h5.close()

if __name__ == "__main__":
    main()
